from typing import Optional, Literal, List


class Quantity:
    def __init__(self, value: Optional[float] = None, comparator: Optional[Literal["<", "<=", ">=", ">"]] = None,
                 unit: Optional[str] = None, system: Optional[str] = "http://unitsofmeasure.org",
                 code: Optional[str] = None):
        """
        :param value: Numerical value (with implicit precision) for Observation
        :param comparator: < | <= | >= | > - how to understand the value
        :param unit: Unit representation
        :param system: System that defines coded unit form
        :param code: Coded form of the unit
        """
        self.value = value
        self.comparator = comparator
        self.unit = unit
        self.system = system
        self.code = code

    def get(self):
        quantity = {
            "value": self.value if isinstance(self.value, float) or isinstance(self.value, int) else None,
            "comparator": self.comparator if self.comparator in ["<", "<=", ">=", ">"] else None,
            "unit": self.unit if isinstance(self.unit, str) else None,
            "system": self.system if isinstance(self.system, str) else None,
            "code": self.code if isinstance(self.code, str) else None
        }
        return {k: v for k, v in quantity.items() if v not in ("", None)}


class Coding:

    def __init__(self, system: str = "", version: str = "", code: str = "", display: Optional[str] = None,
                 user_selected: Optional[bool] = None):
        self.system = system
        self.version = version
        self.code = code
        self.display = display
        self.user_selected = user_selected

    def get(self):
        coding = {
            "system": self.system if isinstance(self.system, str) else None,
            "version": self.version if isinstance(self.version, str) else None,
            "code": self.code if isinstance(self.code, str) else None,
            "display": self.display if isinstance(self.display, str) else None,
            "userSelected": self.user_selected if isinstance(self.user_selected, bool) else None
        }
        return {k: v for k, v in coding.items() if v not in ("", None)}


class CodeableConcept:

    def __init__(self, codings: List[Coding] = None, text: str = ""):
        self.codings = codings
        self.text = text

    def get(self):
        codeableconcept = {
            "coding": [coding.get() for coding in self.codings if isinstance(coding, Coding)] if isinstance(
                self.codings, list) else None,
            "text": self.text if isinstance(self.text, str) else None
        }

        return {k: v for k, v in codeableconcept.items() if v not in ("", None, [])}


class Range:

    def __init__(self, low: Optional[float] = None, high: Optional[float] = None):
        self.low = low
        self.high = high

    def get(self):
        _range = {
            "low": self.low if isinstance(self.low, float) else None,
            "high": self.high if isinstance(self.high, float) else None
        }
        return {k: v for k, v in _range.items() if v not in ("", None)}


class Ratio:

    def __init__(self, numerator: Optional[Quantity] = None, denominator: Optional[Quantity] = None):
        self.numerator = numerator
        self.denominator = denominator

    def get(self):
        ratio = {
            "numerator": self.numerator.get() if isinstance(self.numerator, Quantity) else None,
            "denominator": self.denominator.get() if isinstance(self.denominator, Quantity) else None
        }
        return {k: v for k, v in ratio.items() if v not in ("", None)}


class SampledData:

    def __init__(self, origin: str, period: float, dimensions: int, factor: Optional[float] = None,
                 lower_limit: Optional[float] = None, upper_limit: Optional[float] = None, data: Optional[str] = None):
        self.origin = origin
        self.period = period
        self.dimensions = dimensions
        self.factor = factor
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.data = data

    def get(self):
        sampled_data = {
            "origin": self.origin if isinstance(self.origin, str) else None,
            "period": self.period if isinstance(self.period, float) else None,
            "factor": self.factor if isinstance(self.factor, float) else None,
            "lowerLimit": self.lower_limit if isinstance(self.lower_limit, float) else None,
            "upperLimit": self.upper_limit if isinstance(self.upper_limit, float) else None,
            "dimensions": self.dimensions if isinstance(self.dimensions, int) and self.dimensions > 0 else None,
            "data": self.data if isinstance(self.data, str) else None
        }
        return {k: v for k, v in sampled_data.items() if v not in ("", None)}


class Period:

    def __init__(self, start: str = '', end: str = ''):
        self.start = start
        self.end = end

    def get(self):
        period = {
            "start": self.start if isinstance(self.start, str) else None,
            "end": self.end if isinstance(self.end, str) else None
        }
        return {k: v for k, v in period.items() if v not in ("", None)}


class ObservationValue:

    def __init__(self, value_quantity: Optional[Quantity] = None,
                 value_codeable_concept: Optional[CodeableConcept] = None, value_string: Optional[str] = None,
                 value_boolean: Optional[bool] = None, value_integer: Optional[int] = None,
                 value_range: Optional[Range] = None, value_ratio: Optional[Ratio] = None,
                 value_sampled_data: Optional[SampledData] = None, value_time: Optional[str] = None,
                 value_date_time: Optional[str] = None, value_period: Optional[Period] = None):
        self.value_quantity = value_quantity
        self.value_codeable_concept = value_codeable_concept
        self.value_string = value_string
        self.value_boolean = value_boolean
        self.value_integer = value_integer
        self.value_range = value_range
        self.value_ratio = value_ratio
        self.value_sampled_data = value_sampled_data
        self.value_time = value_time
        self.value_date_time = value_date_time
        self.value_period = value_period

    def get(self):
        value = {
            "valueQuantity": self.value_quantity.get() if isinstance(self.value_quantity, Quantity) else None,
            "valueCodeableConcept": self.value_codeable_concept.get() if isinstance(self.value_codeable_concept,
                                                                                    CodeableConcept) else None,
            "valueString": self.value_string if isinstance(self.value_string, str) else None,
            "valueBoolean": self.value_boolean if isinstance(self.value_boolean, bool) else None,
            "valueInteger": self.value_integer if isinstance(self.value_integer, int) else None,
            "valueRange": self.value_range.get() if isinstance(self.value_range, Range) else None,
            "valueRatio": self.value_ratio.get() if isinstance(self.value_ratio, Ratio) else None,
            "valueSampledData": self.value_sampled_data.get() if isinstance(self.value_sampled_data,
                                                                            SampledData) else None,
            "valueTime": self.value_time if isinstance(self.value_time, str) else None,
            "valueDateTime": self.value_date_time if isinstance(self.value_date_time, str) else None,
            "valuePeriod": self.value_period.get() if isinstance(self.value_period, Period) else None
        }
        return {k: v for k, v in value.items() if v not in ("", None)}



from typing import Optional
from .elements import ObservationValue


class Measurement:
    def __init__(self, value: ObservationValue, code: str, code_system="http://loinc.org",
                 display: Optional[str] = None):

        if not isinstance(value, ObservationValue):
            raise ValueError(f"value={value} is not an ObservationValue type")
        elif not isinstance(code, str):
            raise ValueError(f"code={code} is not an instance of type str")
        elif display is not None and not isinstance(display, str):
            raise ValueError(f"display={display} is not an instance of type str")

        self.value = value
        self.code = code
        self.code_system = code_system
        self.display = display

    def __repr__(self):
        return (f"Measurement(value={self.value}, code='{self.code}', value_system='{self.code_system}')")

    def get(self):
        measurement = {
            "value": self.value.get(),
            "code": self.code,
            "codeSystem": self.code_system,
            "display": self.display if isinstance(self.display, str) else ""
        }
        return {k: v for k, v in measurement.items() if v not in ("", None)}



def is_observation_type(variable):
    if isinstance(variable, dict):
        required_keys = {"value_system", "value", "code", "units_system", "units"}

        if required_keys.issubset(variable.keys()):
            if isinstance(variable["value"], (int, float, str)) and \
                    isinstance(variable["value_system"], str) and \
                    isinstance(variable["code"], str) and \
                    isinstance(variable["units_system"], str) and \
                    isinstance(variable["units"], str):
                return True
    return False



from pathlib import Path
import pydicom
from pydicom.uid import UID
from .terms import SNOMEDCT
from .ehr import Measurement
import json
from concurrent.futures import ThreadPoolExecutor
import os


class Annotator:

    def __init__(self, dataset_path):
        self.descriptions = {}
        self.root = Path(dataset_path)
        self._analysis_dataset()

    def _analysis_dataset(self):
        primary_folder = self.root / "primary"
        if not primary_folder.exists():
            self.descriptions = {}
            raise ValueError(
                'The dataset structure is not based on a SPARC SDS dataset format, please check it and try again!')

        self.descriptions["dataset"] = {
            "id": "",
            "uuid": "",
            "name": self.root.name,
            "path": "/",
        }

        self.descriptions["patients"] = []

        patients_dir = [x for x in primary_folder.iterdir() if x.is_dir()]

        for p in patients_dir:
            patient = {
                "id": "",
                "uuid": "",
                "name": p.name,
                "path": p.relative_to(self.root).as_posix(),
                "observations": [],
                "imagingStudy": self._analysis_dicom_study_samples(p)
            }
            self.descriptions["patients"].append(patient)

    def _analysis_dicom_study_samples(self, study):

        if study.exists():
            imaging_study = {
                "endpointUrl": "",
                "path": study.relative_to(self.root).as_posix(),
                "series": []
            }
            sams = [x for x in study.iterdir() if x.is_dir()]
            if len(sams) < 1:
                return imaging_study

            if len(self.descriptions["patients"]) < 5:
                for sam in sams:
                    s = self._read_sam(sam)
                    if s is not None:
                        imaging_study["series"].append(s)
            else:
                imaging_study["series"] = self._analysis_dicom_samples_worker(sams)

            return imaging_study

        return None

    def _read_sam(self, sam):
        try:
            dcm_files = list(sam.glob("*.dcm"))
            if len(dcm_files) < 1:
                return
            s_dicom_file = pydicom.dcmread(dcm_files[0])
            body_part_examined = s_dicom_file.get((0x0018, 0x0015), None)
            body_site = SNOMEDCT.get(body_part_examined.value.upper(),
                                     None) if body_part_examined is not None else None

            suid = s_dicom_file.get((0x0020, 0x000e), None)
            s = {
                "endpointUrl": "",
                "uid": suid.value if suid is not None else "",
                "name": sam.name,
                "numberOfInstances": len(dcm_files),
                "bodySite": body_site,
                "instances": self._analysis_dicom_sample_instances(dcm_files)
            }
            return s
        except Exception as e:
            print(f"Error reading {sam}: {e}")
            return None

    def _analysis_dicom_samples_worker(self, sams):
        samples = []
        max_workers = os.cpu_count()
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = executor.map(self._read_sam, sams)

        for result in results:
            if result is not None:
                samples.append(result)

        return samples

    @staticmethod
    def _analysis_dicom_sample_instances(dcms):
        instances = []
        for d in dcms:
            dcm = pydicom.dcmread(d)

            # Get the SOP Class UID
            sop_class_uid = dcm.SOPClassUID
            # Get the SOP Class Name using the UID dictionary
            sop_class_name = UID(sop_class_uid).name

            instance = {
                "uid": dcm[(0x0008, 0x0018)].value,
                "sopClassUid": sop_class_uid,
                "sopClassName": sop_class_name,
                "number": dcm[(0x0020, 0x0013)].value
            }
            instances.append(instance)
        return instances

    def add_measurements(self, subjects, measurements):
        self.add_measurement(subjects, measurements)
        return self

    def add_measurement(self, subjects, measurement):
        if isinstance(subjects, list) and isinstance(measurement, list):
            for s in subjects:
                self.add_measurements_by_subject(s, measurement)
        elif isinstance(subjects, list) and isinstance(measurement, Measurement):
            m = measurement
            for s in subjects:
                self.add_measurement_by_subject(s, m)
        elif isinstance(subjects, str) and isinstance(measurement, list):
            s = subjects
            for m in measurement:
                self.add_measurement_by_subject(s, m)
        elif isinstance(subjects, str) and isinstance(measurement, Measurement):
            s = subjects
            m = measurement
            self.add_measurement_by_subject(s, m)
        return self

    def add_measurements_by_subject(self, subject, measurements):
        s = subject
        for m in measurements:
            self.add_measurement_by_subject(s, m)

        return self

    def add_measurement_by_subject(self, subject, measurement):
        if not isinstance(subject, str):
            raise ValueError(f"subject={subject} is not an instance of type str")
        if not isinstance(measurement, Measurement):
            raise ValueError(f"measurement={measurement} is not an instance of type Measurement")
        subject_path = self.root / "primary" / subject
        if not subject_path.exists():
            raise ValueError(f"subject_path={subject_path} is not exists")

        matched_patient = [p for p in self.descriptions.get("patients") if
                           p.get("path") == subject_path.relative_to(self.root).as_posix()][0]
        matched_patient["observations"].append(measurement.get())

        return self

    def save(self, path=None):
        if path:
            save_path = Path(path) / "measurements.json"
        else:
            save_path = self.root / "measurements.json"

        with open(save_path, "w") as json_file:
            json.dump(self.descriptions, json_file, indent=4)
